// Copyright (c) 2021 The Chromium OS Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
#ifndef PERF_DATA_CONVENTER_ARM_SPE_DECODER
#define PERF_DATA_CONVENTER_ARM_SPE_DECODER

#include <map>
#include <sstream>
#include <cstdint>
#include <string>

#ifndef BIT
#define BIT(n) (1UL << (n))
#endif

/*
  Example of supporting Arm SPE profile packet:

  1. load/str/other instruction
    b0 68 ad d0 0b fc ff 00 80                      PC 0xfffc0bd0ad68 el0 ns=1
    99 0a 00                                        LAT 10 ISSUE
    98 c1 00                                        LAT 193 TOT
    52 1e 06                                        EV RETIRED L1D-ACCESS L1D-REFILL TLB-ACCESS LLC-REFILL REMOTE-ACCESS
    49 00                                           LD
    b2 d8 d6 b6 0b fc ff 00 00                      VA 0xfffc0bb6d6d8
    9a 01 00                                        LAT 1 XLAT
    9e b6 00                                        LAT 182
    00 00 00 00                                     PAD
    01                                              END

  2. branch instruction
    b0 20 a7 d0 0b fc ff 00 80                      PC 0xfffc0bd0a720 el0 ns=1
    99 66 00                                        LAT 102 ISSUE
    98 67 00                                        LAT 103 TOT
    52 82 00                                        EV RETIRED MISPRED
    4a 01                                           B COND
    b1 a8 ad d0 0b fc ff 00 80                      TGT 0xfffc0bd0ada8 el0 ns=1
    00 00 00 00 00 00                               PAD
    01                                              END
*/

namespace quipper {
/* Use bit state to represent Arm SPE types.
   The bit status definition is the same as ArmSpeRawType */
using ArmSpeEventType = uint16_t;

/* DO NOT CHANGE IT, it`s define at kernel/tool/perf/.../arm_spe_decoder.h */
enum ArmSpeRawType {
  EV_EXCEPTION_GEN = 0, // not used
  EV_RETIRED = 1,
  EV_L1D_ACCESS = 2,    // not used
  EV_L1D_REFILL = 3,
  EV_TLB_ACCESS = 4,    // not used
  EV_TLB_REFILL = 5,    // not used
  EV_NOT_TAKEN = 6,     // not used
  EV_MISPRED = 7,       // not used
  EV_LLC_ACCESS = 8,    // not used
  EV_LLC_REFILL = 9,
  EV_REMOTE_ACCESS = 10,// not used
};

/* Define instruction type of Arm SPE packet */
enum class ArmSpeInstType {
  ARM_SPE_INST_OTHER,     // other instruction
  ARM_SPE_INST_LD_ST,     // ld/str instruction
  ARM_SPE_INST_BR,        // branch instruction
  ARM_SPE_INST_BR_COND,   // branch condition
  ARM_SPE_INST_BR_IND,    // branch indirect call
};

inline bool IsArmSpeBranchInst(const ArmSpeInstType& type)
{
  if (type == ArmSpeInstType::ARM_SPE_INST_BR || type == ArmSpeInstType::ARM_SPE_INST_BR_COND
   || type == ArmSpeInstType::ARM_SPE_INST_BR_IND) {
    return true;
  }
  return false;
}

struct ArmSpeParsedPacket {
  uint64_t sampleAddr = 0;        // Address of sample instruction
  uint64_t targetAddr = 0;        // Branch instrution jump target address
  ArmSpeEventType packetType = 0; // One packet may have multiply types, it represent as bit
  ArmSpeInstType instType;        // Instruction Type
};

class ArmSpePacketDecoder {
public:
  explicit ArmSpePacketDecoder() {
    InitProcessArmSpePacketMap();
  }
  ~ArmSpePacketDecoder() {}

  bool Init(const std::string* traceData);

  bool IsUnsolvedPacketExist() const;

  bool SolveOnePacket(ArmSpeParsedPacket& packet);

  void HandleError();

private:
  enum class ArmParserState {
    START,
    PC,
    LAT,
    TYPE,
    LD_ST,
    BRANCH,
    TGT,
    INST_OTHER,
    PAD,
    END,
    EXIT,
    UNKNOWN,
  };

  enum ArmSpeInstTag {
    PC = 0xb0,
    END = 0x01,
    PAD = 0x00,
    TYPE = 0x52,
    LD_ST = 0x49,
    BRANCH = 0x4a,
    TGT = 0xb1,
    INST_OTHER = 0x48,
  };

  enum ArmSpeBrTypeTag {
    B = 0x00,       // branch instruction
    B_COND = 0x01,  // condition instruction
    B_IND = 0x02,   // indirect branch instruction
  };

  /* Init Table-driven map to parser a SPE packet. */
  void InitProcessArmSpePacketMap();

  bool ProcessInitSection(ArmSpeParsedPacket& packet);

  bool ProcessPCSection(ArmSpeParsedPacket& packet);

  bool ProcessLATSection(ArmSpeParsedPacket& packet);

  bool ProcessTypeSection(ArmSpeParsedPacket& packet);

  bool ProcessLDSTSection(ArmSpeParsedPacket& packet);

  bool ProcessBranchTGTSection(ArmSpeParsedPacket& packet);

  bool ProcessInstOtherSection(ArmSpeParsedPacket& packet);

  bool ProcessPadSection(ArmSpeParsedPacket& packet);

  bool ProcessEndSection(ArmSpeParsedPacket& packet);

  std::string GetArmParserStateString(ArmParserState);

  const char* m_rawTraceData = nullptr;
  size_t m_totalSize = 0;
  size_t m_curPos = 0;
  enum ArmParserState m_curParserState;
  ArmSpeInstType m_curInstType;

  /* Table-Driven approach to parser a SPE packet. */
  std::map<ArmParserState, bool (ArmSpePacketDecoder::*)(ArmSpeParsedPacket&)> m_decodeArmSpePacketMap;

}; // class ArmSpePacketDecoder

} // namespace quipper

#endif  //PERF_DATA_CONVENTER_ARM_SPE_DECODER
